package adt.service;

import adt.client.MainNodeClient;
import adt.entity.Port;
import adt.entity.ScanInfo;
import adt.entity.Vlan;
import adt.dto.Node;
import adt.repository.PortRepository;
import adt.repository.ScanInfoRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;


import java.util.concurrent.CountDownLatch;

import java.io.IOException;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import java.util.stream.Stream;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ExecutionException;
import java.lang.InterruptedException;
import java.lang.Thread;
import lombok.extern.slf4j.Slf4j;

@Service
@Slf4j
public class ScheduleService {

    private final MainNodeClient mainNodeClient;
    private final HostScannerService hostScannerService;
    private final PortScannerService portScannerService;
    private final SendAssetService sendAssetService;

    private final ScanInfoRepository scanInfoRepository;
    private final PortRepository portRepository;

    private List<Vlan> vlan;
    private Node node;

    private ScheduledExecutorService scheduledExecutorService;
    private ScheduledFuture<?> futureTask;
    private Runnable task;

    private Long prevSystemDelay;
    private Long iniExecutionTime;


    private class ScheduledTask implements Runnable {

        @Override
        public void run() {

            log.info("NEW EXECUTION SCHEDULED. Number of Vlans to scan: {}", vlan.size());
            iniExecutionTime = System.currentTimeMillis();

            CountDownLatch latch = new CountDownLatch(vlan.size());
            ThreadPoolExecutor executor = (ThreadPoolExecutor) Executors.newFixedThreadPool(vlan.size());

            // Timestamp ini
            Date iniTimestamp_scan = new Date();

            vlan.forEach(element -> executor.submit(() -> {

                        log.info("Vlan address ready to scan: " + element.getNetAddress());

                        hostScannerService.scanNetwork(element.getVlanName(), element.getNetAddress(), element.getMask());
                        log.info("SCANNING OF NETWORK {} SUCCESSFULLY COMPLETED", element.getNetAddress());

                        Date endTimestamp_scan = new Date();

                        ScanInfo scanInfo  = ScanInfo.createScanInfo(element.getVlanName(), node.getIp(), iniTimestamp_scan, endTimestamp_scan, "scan");

                        scanInfoRepository.save(scanInfo);

                        latch.countDown();

            }));

            try {
                latch.await();
            } catch (InterruptedException E) {
                // handle
            }
            executor.shutdown();

            List<Port> ports = portScannerService.getPortList();

            portRepository.saveAll(ports);

            //boolean iniFlag = false;

            log.info("All networks have been successfully scanned and its assets detected. Now managing this information...\n");
            Long systemDelay = sendAssetService.sendAssets(false, vlan);//, "scan");

            boolean flag = true;
            if(systemDelay == null){
                flag = false;
            }else {
                flag = !Long.toString(systemDelay).equals(Long.toString(vlan.get(0).getDelay()));
            }
            if( flag ){

                futureTask.cancel(false);
                prevSystemDelay = vlan.get(0).getDelay();


                log.info("ATTENTION, a new execution delay has been set for this scanning process.\n--Previous: {} seconds.\n--New: {} seconds.\nNext iterations will use this new delay.\n", vlan.get(0).getDelay(), systemDelay);

                vlan.get(0).setDelay(systemDelay); // PONERSELO A TODOS

                /* At this point in the execution, it is guaranteed that the futureTask's thread has finished, so it is safely cancelled and we can schedule a new futureTask. */
                scheduleSetUp(node);
            }

        }
    }

    @Autowired
    public ScheduleService(HostScannerService hostScannerService, PortRepository portRepository,PortScannerService portScannerService, SendAssetService sendAssetService, MainNodeClient mainNodeClient, ScanInfoRepository scanInfoRepository) {
        this.mainNodeClient = mainNodeClient;
        this.hostScannerService = hostScannerService;
        this.sendAssetService = sendAssetService;
        this.portRepository = portRepository;
        this.portScannerService = portScannerService;
        this.scanInfoRepository = scanInfoRepository;
        scheduledExecutorService = Executors.newSingleThreadScheduledExecutor();

        task = new ScheduledTask();
    }

    void scheduleSetUp(Node node) {

        if (this.vlan == null){
            this.node = node;
            this.vlan = node.getVlan();
        }

        if(this.prevSystemDelay != null){

            Long diff = (System.currentTimeMillis() - this.iniExecutionTime) / 1000;
            if( (diff) < this.prevSystemDelay ){

                Long remaining = (this.prevSystemDelay - diff) * 1000;
                log.info("Stalling execution until a delay of " + this.prevSystemDelay + " seconds between tasks is satisfied.\n--Time already covered: " + diff + " seconds.\n--Remaining: " + (remaining/1000) + " seconds.");
                log.info("Stalling execution until a delay of {} seconds between tasks is satisfied.\n--Time already covered: {} seconds.\n--Remaining: {} seconds.",  this.prevSystemDelay, diff, (remaining/1000));

                try{
                    Thread.sleep(remaining);
                }catch(InterruptedException ie){
                    // Handled exception
                }

            }

            this.prevSystemDelay = null;
        }

        log.info("VLAN DELAY INFO: {} seconds.", vlan.get(0).getDelay());


        futureTask = scheduledExecutorService.scheduleAtFixedRate(task, 0, vlan.get(0).getDelay(), TimeUnit.SECONDS);
    }

}
